
from __future__ import print_function

import argparse as ap
import getpass
import os
import xml.etree.ElementTree as ET
from pathlib import Path

import sys
from pyat import version
from pyat.pymygdala import RPCEvent
from pycapo import CapoConfig

# Help Messages
_PASS_DESCRIPTION = """AAT/PPI QA Pass {}: The calibration jobs listed are considered acceptable.  This tool initiates
 the ingestion of calibration products for the indicated directories and updates the list of calibrations for the 
 associated execution blocks."""

_FAIL_DESCRIPTION = """AAT/PPI QA Fail {}: This tool marks the execution blocks related to the indicated directories 
as not appropriate for automated calibration, and initiates a clean up procedure to remove the directories from the QA 
area. """

# Pipeline Processing Request
PPR = "PPR"
PPR_FILENAME = PPR + ".xml"


def _make_parser(status):
    r""" Build a command line parser for this app: this is external to MrClean and/or CapoSettings because both need
    it, and Sphinx will want a look at ot to build docs. """

    if status == "PASS":
        _DESCRIPTION = _PASS_DESCRIPTION
    else:
        _DESCRIPTION = _FAIL_DESCRIPTION

    result = ap.ArgumentParser(description=_DESCRIPTION.format(version),
                               formatter_class=ap.RawTextHelpFormatter)
    result.add_argument('-P', '--profile', action='store', default='',
                        help='profile name to use, e.g. test, production')
    result.add_argument('-f', '--fileSetId', action='store', default='', help='provide a fileSetId for the directory instead of searching for it')
    # For capturing DA comments, eventually.
    # result.add_argument('-c', '--comment', type=str, action='store', default='No Comment Provided', help='Comment on the EB.')
    result.add_argument('-e', '--email', type=str, action='store', default='', help='Email for completion notification')
    result.add_argument('directoryName', type = str, action='store', nargs='+', help='name of the qa2 subdirectory to which the status applies')
    return result


def rawdata_check(path):
    #TODO: Integrate this with both searches
    # If we arrived here, then we need another route:
    #
    # the rawdata is stored in a directory named for
    # fileSetId
    raw_path = Path(path)
    fileList = list(raw_path.glob('**/*'))
    if(1 < len(fileList)):
        # This lists:  [fsid, ASDMBinary] if successful
        fileSetId = fileList[0].name.replace("/", "") # make sure no trailing slash
        return fileSetId


def qa_search(path, subdir):
    r"""Seach the /qa2/ subdirectory for the listed directory and extract the fileSetId from the
    name of the flagversions file.  return an empty string if this search fails"""
    fileSetId = ''


    # Where we're looking
    qaPath = Path(path + '/' + subdir + '/products/')

    # There's quite a bit of junk in the directory, and we only need the one tgz file
    # so screen out the rest of them
    #
    # Loosened the glob to allow non NN[ABCD]-NNN.sb... identifiers through
    fileList = list(qaPath.glob('**/*.ms.flagversions.tgz'))
    if(0 < len(fileList)):
        flagTableFile = fileList[0].name
        fileSetId = flagTableFile.split(".ms")[0]
        return fileSetId

    #
    # If we arrived here, then we need another route:
    #
    # the rawdata is stored in a directory named for
    # fileSetId
    raw_path = Path(path + '/' + subdir + '/rawdata/')
    fileList = list(raw_path.glob('**/'))
    if(1 < len(fileList)):
        # This lists:  [rawdata, fsid, ASDMBinary] if successful
        fileSetId = fileList[1].name.replace("/", "") # make sure no trailing slash
        return fileSetId

    #
    # Ok, now things are getting weird.
    #
    return fileSetId


def spool_search(path, subdir):
    r"""Search the /spool/ for the desired directory, check for
    a rawdata subdirectory and parse the PPR to discover the fileSetId.
    return an empty string if you can't acquire the fileSetId"""
    fileSetId = ''

    # If we're here, it's because there's nothing useful in the qa2
    # directory, which has most of the data in it....
    #

    # Try the rawdata directory again
    #
    #
    # the rawdata is stored in a directory named for
    # fileSetId
    raw_path = Path(path + '/' + subdir + '/rawdata/')
    fileList = list(raw_path.glob('**/'))
    if(1 < len(fileList)):
        # This lists:  [rawdata, fsid, ASDMBinary] if successful
        fileSetId = fileList[1].name.replace("/", "") # make sure no trailing slash
        return fileSetId

    # No luck with the raw data directory here in spool...
    #
    # So that leaves us with one last thing to try:  PPR.xml

    pprFile = path+'/'+subdir+'/working/' + PPR_FILENAME
    try:
        fileTree = ET.parse(pprFile)
        #
        # Now walk the structure of the PPR file down to the piece we need
        #
        procRequests = fileTree.find('ProcessingRequests')
        procReq = procRequests.find('ProcessingRequest')
        dataSet = procReq.find('DataSet')
        sdmElement = dataSet.find('SdmIdentifier').text
        if sdmElement is not None:
            fileSetId = sdmElement
            return fileSetId

    except ET.ParseError as pe:
        print("Error parsing PPR file.")
    except StopIteration as si:
        print('No SdmIdentifier in the PPR file.')
    except FileNotFoundError as fnf:
        print('No PPR file found.')

    # We haven't managed to find the fileSetId: indicate that
    return ''


def qa_list(status):
    r"""From the given directory(s) find the fileSetId(s) required for either workflow or database update
    purposes (either from some filenames in the directory, or potentially parsing the PPR if that is deemed
    necessary). """
    # Get the command line arguments
    args = _make_parser(status).parse_args()

    # Shamelessly stolen from epilogue with a twist: allow for explict profile setting via the CL
    if 'CAPO_PROFILE' not in os.environ and '' == args.profile:
        # try to synthesize a profile from our installation root
        profile = os.path.abspath(sys.argv[0]).split(os.path.sep)[-3]
        os.environ['CAPO_PROFILE'] = profile
        print('No CAPO_PROFILE explicitly set, synthesizing {0} by path'.format(str(profile)))
    elif '' != args.profile:
        os.environ['CAPO_PROFILE'] = args.profile

    capo = CapoConfig()

    qaDirectory    = capo.getstring("edu.nrao.archive.workflow.config.CiplWorkflowSettings.qaDirectory")
    spoolDirectory = capo.getstring("edu.nrao.archive.workflow.config.CiplWorkflowSettings.spoolDirectory")


    #
    # So, the fileSetId is both in the name of a secondary tar file in the QA directory,
    # and in the casa PPR file (which lives in spoolDirectory/$subDirName/working/)
    # The former is by far the easier to deal with.
    #
    # A QA products directory has the following tgz files
    #
    # 13B-326.sb29065738.eb29113188.56764.86226100694.ms.flagversions.tgz
    # unknown.session_1.caltables.tgz
    # weblog.tgz
    #
    # The first of which is the fileSetId, so for each directory we were given,
    # parse the fileSetId and send an event for each.
    #
    for resultsDirectory in args.directoryName:
        # consistency: we'll place the /s
        resultsDirectory = resultsDirectory.replace("/", "")

        # If we haven't been provided with the fileSetId, go find it
        if('' == args.fileSetId):
            fileSetId = qa_search(qaDirectory, resultsDirectory)
            if('' == fileSetId):
                fileSetId = spool_search(spoolDirectory, resultsDirectory)
                if ('' == fileSetId):
                    print("Failed to obtain a fileSetId for directory " + resultsDirectory)
                    return
        else:
            fileSetId = args.fileSetId

        # We've got the information we need, go ahead and send the event
        print("Sending "+status+" for "+resultsDirectory+" ("+fileSetId+")")
        qa_single(fileSetId, resultsDirectory, status, args.email)


def qa_single(fileset_id, results_directory, status, notify_email):
    r"""Takes two strings: fileSetId (parsed from a file in the directory we're interested in)
    and status (Pass or Fail), and builds string to execute to create the appropriate event
    for updating the database and/or launching a workflow."""

    # A little parsing work:  The project code is the first part of the fileSetId
    project_code = fileset_id.split('.')[0]

    # We'll want to know who sent the message:
    analyst_name = getpass.getuser()

    # now prep the event, use an RPC event to get data back from the system:
    broadcaster = RPCEvent(profile=os.environ['CAPO_PROFILE'], exchange='archive.events', application='qa-script')

    event_data = {
        "user": analyst_name,
        "message": "{}={}".format(fileset_id, status),
        "logData": {
            "projectCode": project_code,
            "status": status,
            "type": "calibration",
            "fileset_id": fileset_id,
            "markedDirectory": results_directory
        }
    }

    if '' != notify_email:
        event_data["logData"]["email"] = notify_email

    broadcaster.send(event_data)

    # Now get the reply: (uncomment when that part is done)
    ingestion_data = broadcaster.get_reply()


    directory = None
    log_id = None

    if "working_directory" in ingestion_data:
        directory = ingestion_data["working_directory"]

    if "logging_id" in ingestion_data:
        log_id = ingestion_data["logging_id"]

    if "PASS" == status:
        workflow_name = 'Calibration Ingestion'
    else:
        workflow_name = 'Qa Cleanup'

    print("{} ({}) is running in {}".format(workflow_name, log_id, directory))

    return


def qa_pass():
    qa_list("PASS")


def qa_fail():
    qa_list("FAIL")
